 /*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: boost-r.c 543 2017-11-17 17:14:27Z gouzy $
 * 
 ---------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/sysinfo.h>
#include "boost-r.h"
#include <sys/time.h>
              

int main(int argc,char *argv[])
{
	pid_t pid = getpid();
	char tmpstr[MAXLINE];
	tmpstr[0]='\0';
	paramparser_t *o_param=NULL;
	read_booster_t    read_booster;

	char auth[500] = "";
	unsigned long long max_mem_usage=0;
	
	char rev[500];
	sprintf(rev,"$Rev: 543 $ Compiled on %s at %s, AUTH=%zu GCC=%d.%d.%d\n", __DATE__, __TIME__,strlen(auth),__GNUC__,__GNUC_MINOR__,__GNUC_PATCHLEVEL__);
	fprintf(stderr,"developed by jerome.gouzy@inra.fr\n");
	fprintf(stderr,"uthash:     Copyright (c) 2005-2014, Troy D. Hanson  http://troydhanson.github.com/uthash/ All rights reserved.\n");
	fprintf(stderr,"openssl:    Copyright (c) 1998-2011 The OpenSSL Project. All rights reserved. http://www.openssl.org/source/license.html\n");
	
	ParamParserPrintCli(argc,argv,stderr,rev);
	
#ifdef REVMAP
	fprintf(stderr,"REVMAP mode for large datasets (warning: less sensitive)\n");
#endif
	
	if ( argc == 1 )
	{
		Usage(argv);
		exit(0);
	}
	
	SetStackSize(100000);
	
	sprintf(tmpstr,"boost-r.time.%d",(int)pid);
	
	char param_str[5000] = "help short=s interleaved outdir=s klen=i minkmers=i maxkmers=i minkmers_for_lr_extension=i long=s \
	nb_threads=i size=s depth=i phix build_long_reads=i steps=s insert_size=i \
	preprocessed_reads=s preprocessed_LR=s templ_depth=i minovl=i minpci=s maxovh=i build_very_long_reads=i maxmm=i compress=i save_reads \
	minkdepth=i maxpaircov=i dump_klib=s load_klib=s stranded_rnaseq illumina_adapt=s DEBUG";
	ParamParserInit(&o_param,argc,argv,param_str);
	
	if ( ParamParserIsDefined(&o_param,"preprocessed_LR") )
	{
		ParamParserAssertDefined(&o_param,"preprocessed_reads");
		ParamParserSetUnlessDefined(&o_param,"steps","");
		ParamParserSetUnlessDefined(&o_param,"build_long_reads","0");
	}
	else
	{
		ParamParserSetUnlessDefined(&o_param,"build_long_reads","1");
		if ( ! ParamParserIsDefined(&o_param,"preprocessed_reads") )
			ParamParserSetUnlessDefined(&o_param,"steps","01234");
	}
	ParamParserSetUnlessDefined(&o_param,"build_very_long_reads","1");
	
//	ParamParserSetUnlessDefined(&o_param,"steps","12");
	ParamParserSetUnlessDefined(&o_param,"klen",itoa(DEFAULT_PARAM_KLEN,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"nb_threads",itoa(DEFAULT_PARAM_NB_THREADS,tmpstr));
//  ParamParserSetUnlessDefined(&o_param,"stranded_rnaseq",itoa(FALSE,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"outdir",".");
	ParamParserSetUnlessDefined(&o_param,"insert_size",itoa(DEFAULT_PARAM_INSERT_SIZE,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"depth",itoa(DEFAULT_PARAM_DEPTH,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"templ_depth",ParamParserGet(&o_param,"depth"));
	ParamParserSetUnlessDefined(&o_param,"maxmm",itoa(DEFAULT_PARAM_MAXMM,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"minovl",itoa(DEFAULT_PARAM_MINOVL,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"minpci",ftoa(DEFAULT_PARAM_MINPCI,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"maxovh",itoa(DEFAULT_PARAM_MAXOVH,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"minkdepth",itoa(DEFAULT_PARAM_MINKDEPTH,tmpstr));
//	ParamParserSetUnlessDefined(&o_param,"maxpaircov",ParamParserGet(&o_param,"depth"));
	ParamParserSetUnlessDefined(&o_param,"maxpaircov",itoa(DEFAULT_PARAM_MAXPAIRCOV,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"compress",itoa(DEFAULT_PARAM_COMPRESS,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"adapter",ILLUMINA_ADAPT);
	ParamParserSetUnlessDefined(&o_param,"minkmers",itoa(DEFAULT_PARAM_MINKMERS,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"maxkmers",itoa(DEFAULT_PARAM_MAXKMERS,tmpstr));
	ParamParserSetUnlessDefined(&o_param,"minkmers_for_lr_extension",itoa(DEFAULT_PARAM_MINKMERS_FOR_LR_EXTENSION,tmpstr));
	char phix[10000]={0};
	if ( ! ParamParserIsDefined(&o_param,"phix") )
		sprintf(phix,"%s%s",PHIX1,PHIX2);
	ParamParserSetUnlessDefined(&o_param,"phix",phix);
	
	ParamParserAssertDefined(&o_param,"size");
	ParamParserAssertDirExists(&o_param,"outdir");
	
	
	Date("Init");
	
   // ParamParserDump(o_param,stdout);
	
	InitBoostr(&read_booster,o_param);
	read_booster.nb_max_seq= (read_booster.expected_nt_size*read_booster.depth)/(read_booster.min_readlen_foruse*2);
	read_booster.nb_max_seq *= 2.5; // pour prendre de la marge et pouvoir faire le replace, il faut aussi la place pour les vlr mais elle sont 4 * plus longues donc 4 moins nombreuses
	// pour 3.5Gb le max devrait etre de 972222222
	assert(read_booster.nb_max_seq<UINT_MAX);
	fprintf(stderr,"nb_max_seq=%u\n",read_booster.nb_max_seq);
	MallocSeqSet(&read_booster.h_seqcontigs,read_booster.nb_max_seq);
			
	char skeleton[MAXFILENAME]={0};
	char tmpfile[MAXFILENAME]={0};
	char *outdir               = ParamParserGet(&o_param,"outdir");
	int depth                  = atoi(ParamParserGet(&o_param,"depth"));
	if ( ! ParamParserIsDefined(&o_param,"long") )
	{
		sprintf(skeleton,"boost-r.depth%d",depth);
		ParamParserSetUnlessDefined(&o_param,"long",skeleton);
	}
	char *outprefix            = ParamParserGet(&o_param,"long");
	sprintf(skeleton,"%s/%s",outdir,outprefix);
	char fastafile[MAXFILENAME]="";
	char longreadsfile[MAXFILENAME]="";
	char cleanupdataset[MAXFILENAME]="";
	char verylongreadsfile[MAXFILENAME]="";
	
	int templ_depth            = atoi(ParamParserGet(&o_param,"templ_depth"));
	
	char *param_preprocessed_reads       = ParamParserGet(&o_param,"preprocessed_reads");
	if ( param_preprocessed_reads != NULL )
		ParamParserAssertFileExists(&o_param,"preprocessed_reads");
	
	char *param_preprocessed_lr       = ParamParserGet(&o_param,"preprocessed_LR");
	if ( param_preprocessed_lr != NULL )
		ParamParserAssertFileExists(&o_param,"preprocessed_LR");
	
	/*
	char *param_long_reads         = ParamParserGet(&o_param,"file_of_long_reads");
	if ( param_long_reads != NULL )
		ParamParserAssertFileExists(&o_param,"file_of_long_reads");
	*/
	
	char *param_filename_save_klib = ParamParserGet(&o_param,"dump_klib");
	char *param_filename_load_klib = ParamParserGet(&o_param,"load_klib");
	if ( param_filename_load_klib != NULL )
		ParamParserAssertFileExists(&o_param,"load_klib");
		
	char *steps                    = ParamParserGet(&o_param,"steps");
	
	int flag_build_long_reads      = atoi(ParamParserGet(&o_param,"build_long_reads"));
	int flag_build_very_long_reads = atoi(ParamParserGet(&o_param,"build_very_long_reads"));
	
	seqcount_t *h_seqfile2counts=NULL;
	lof_t lof;
	ParamParserInitLof(lof);
	lof_t lof_clean;
	ParamParserInitLof(lof_clean);
	int i=0;
	char ext[1000]="";
	if ( read_booster.compress )
		strcpy(ext,".gz");

	if( read_booster.debug )
	{
		InitMaxMemUsage(&max_mem_usage);
		fprintf(stderr,"Max Memory Usage POST INIT: %d Mb\n",(int)max_mem_usage/1024);
	}
	
	if ( steps != NULL && strlen(steps) )
	{
    	ParamParserAssertDefined(&o_param,"size");
		ParamParserAssertDefined(&o_param,"short");
		ParamParserGetListOfFiles(&o_param,"short",lof);
		AssertDataSets(&read_booster,lof);
	}
	
	if ( param_filename_load_klib != NULL )
	{
		LoadKlib(&read_booster,param_filename_load_klib);
	}
	else if ( param_preprocessed_reads == NULL )
	{
		if( read_booster.debug )
			Date("KLIB");
		BuildKmerLibrary(&read_booster,lof,&h_seqfile2counts,FALSE,FALSE,0);
		if( read_booster.debug )
			Date("KLIB done");
	}
	if ( steps != NULL && strchr(steps,'0') ) // clean up dataset
	{
		Date("CLEANUP DATASET");
		int i;
		if ( read_booster.nb_cpus > 1 )
		{
			for(i=0;i<read_booster.nb_cpus;i++)
			{
				sprintf(tmpfile,"%s.cln.%d.fasta%s",skeleton,i+1,ext);
				lof_clean[i]=(char *)malloc(sizeof(tmpfile)+1);
				strcpy(lof_clean[i],tmpfile);
			}
			lof_clean[i]=NULL;
		}
		else
		{
			sprintf(tmpfile,"%s.cln.fasta%s",skeleton,ext);
			lof_clean[0]=(char *)malloc(sizeof(tmpfile)+1);
			strcpy(lof_clean[0],tmpfile);
		}
		sprintf(cleanupdataset,"%s.cln.fasta%s",skeleton,ext);
		CleanUpDataset(&read_booster,lof,lof_clean,cleanupdataset,ParamParserGet(&o_param,"adapter"),ParamParserGet(&o_param,"phix"),&h_seqfile2counts);
		ResetReadFileAndKlib(&read_booster,&o_param,cleanupdataset,lof,TRUE,&h_seqfile2counts,FALSE);
		CleanUpDatasetMore(&read_booster,cleanupdataset);
	}
	if ( param_filename_save_klib != NULL )
	{
		Date("DUMPING KLIB");
		SaveKlib(&read_booster,param_filename_save_klib);
	}
	
	if ( param_preprocessed_reads != NULL )
	{
		strcpy(fastafile,param_preprocessed_reads);
		Load(&read_booster,fastafile,FALSE,TRUE,TRUE,NULL,PAIRS);
	}
	if ( param_preprocessed_lr != NULL )
	{
		Load(&read_booster,param_preprocessed_lr,FALSE,TRUE,FALSE,NULL,LR);
	}
	if( read_booster.debug )
	{
		InitMaxMemUsage(&max_mem_usage);
		fprintf(stderr,"Max Memory Usage: %d Mb POST LOAD/CLN\n",(int)max_mem_usage/1024);
	}
	for(i=0;steps != NULL  && i<strlen(steps);i++)
	{
		switch(steps[i])
		{
			case '0':
				break;
			case '1':
				Date("STEP1");
				ParseReads(&read_booster,lof,&h_seqfile2counts,ONLY_UNMAPPED,templ_depth);
				sprintf(fastafile,"%s.%d.sk.fa",skeleton,ONLY_UNMAPPED);
				break;
			case '2':
				Date("STEP2");
				if ( param_preprocessed_reads != NULL )
					ResetKmerize(&read_booster,PAIRS);
				ParseReads(&read_booster,lof,&h_seqfile2counts,INCREASE_DEPTH,depth);
				sprintf(fastafile,"%s.%d.dp.fa",skeleton,INCREASE_DEPTH);
				break;
			case '3':
				if ( read_booster.a_kmercode == NULL )
				{
					fprintf(stderr,"STEP3 mode skipped, it requires a klib\n");
					exit(EXIT_FAILURE);
				}
				Date("PRE STEP3");
				ResetKmerize(&read_booster,PAIRS);
				Date("STEP3");
				// plus stringent que le rescue if ( IdentifyFrequentsButUncapturedKmers(&read_booster) )
#ifdef REVMAP
				BuildKmerLibraryFromSelection(&read_booster,FALSE);
#endif
				ParseReads(&read_booster,lof,&h_seqfile2counts,RESCUE_MISSING_KMERS,depth);
				sprintf(fastafile,"%s.%d.fx.fa",skeleton,RESCUE_MISSING_KMERS);
				break;
			case '4':
				/* je dois le faire car INCREASE_DEPTH ne le fais pas et EvalKmerDepth va enlever des choses importantes */
				Date("PRE STEP4");
				// 20140520: desormais j'utilise la meme structure de donnee
				// 1: je refais l'inventaire
				ResetKmerize(&read_booster,PAIRS);
				Date("STEP4");
				// 2: je change le status des sequences obsoletes
#ifdef REVMAP
				BuildKmerLibraryFromSelection(&read_booster,FALSE);
				IdentifyErroneousSequencesSeqBySeq(&read_booster);
#else
				IdentifyErroneousSequences(&read_booster);
#endif
				// 3: j'indexe uniquement les sequences obsoletes
				ResetKmerize(&read_booster,OBSOLETE);
				// 4: remplacement des paires jugees comme problematiques (= avec kmer rares) */
				ParseReads(&read_booster,lof,&h_seqfile2counts,REPLACE_PROBLEMATIC_PAIRS,depth);
				// 5: je supprime les sequences obsoletes de la structure de donnees
				CleanUpBoostr(&read_booster,CONTIGER_CLEAN_SEQOBS);
				// 6: je reconstruie la table des kmers avec les bons
				ResetKmerize(&read_booster,PAIRS);
				sprintf(fastafile,"%s.%d.rd.fa",skeleton,REPLACE_PROBLEMATIC_PAIRS);
#ifdef REVMAP
				BuildKmerLibraryFromSelection(&read_booster,FALSE);
#endif
				break;
			case '8':
				Date("SKELETON PAIRS");
				ParseReads(&read_booster,lof,&h_seqfile2counts,ONLY_UNMAPPED_PAIRS,depth);
				sprintf(fastafile,"%s.%d.skel_pairs.fa",skeleton,ONLY_UNMAPPED_PAIRS);
				break;
			default:
				fprintf(stderr,"Error: unmanaged step code\n");
				exit(EXIT_FAILURE);
		}
		if ( ((ParamParserIsDefined(&o_param,"save_reads") && steps[i]!='0') || steps[i] == '4' ))
		{
			// je ne peux pas faire une version threade car LR change le status
			Date("DUMP FASTA");
			DumpFasta(&read_booster,fastafile,TRUE,FALSE,PAIRS,0);
		}
		if( read_booster.debug )
		{
			InitMaxMemUsage(&max_mem_usage);
			fprintf(stderr,"Max Memory Usage: %d Mb POST STEP %c\n",(int)max_mem_usage/1024,steps[i]);
		}
	}
	if ( steps != NULL && read_booster.a_kmercode != NULL)
	{
		if ( 0 && read_booster.debug)
			IdentifyFrequentsButUncapturedKmers(&read_booster);
		CleanUpBoostr(&read_booster,CONTIGER_CLEAN_TMP); /* nettoie les structures de donnees non necessaires */
	}
	if ( strlen(fastafile) > 0 )
		ResetReadFileAndKlib(&read_booster,&o_param,fastafile,lof,TRUE,&h_seqfile2counts,TRUE); // ici je suis sur des lectures nettoyees et je fais le revcomp
	
	if ( param_filename_save_klib != NULL ) // je ne conserve que le fichier de kmer final
	{
		Date("DUMPING UPDATED KLIB");
		SaveKlib(&read_booster,param_filename_save_klib);
	}
	if( read_booster.debug )
	{
		InitMaxMemUsage(&max_mem_usage);
		fprintf(stderr,"Max Memory Usage: %d Mb POST KLIB\n",(int)max_mem_usage/1024);
	}
	
	read_booster.minkdepth = 1; // une fois que les reads sont nettoyes, meme si je ne vois qu'une fois c'est bon
					
	if ( flag_build_long_reads )
	{
		Date("RK LR");
		ResetKmerize(&read_booster,PAIRS);
		Date("LR");
		sprintf(longreadsfile,"%s.LR.fa",skeleton);
		BuildLongReads(&read_booster,longreadsfile,&max_mem_usage);
	}
	if( read_booster.debug )
	{
		InitMaxMemUsage(&max_mem_usage);
		fprintf(stderr,"Max Memory Usage: %d Mb POST LR\n",(int)max_mem_usage/1024);
	}
		
	if ( flag_build_very_long_reads )
	{
		// changement, les lr doivent etre en ram
		Date("RK VLR");
		ResetKmerize(&read_booster,LR);
		Date("VLR");
		sprintf(verylongreadsfile,"%s.VLR.fa",skeleton);
		read_booster.nrdb=FALSE; // je desactive, c'est trop long et la suppression de la redondance gene cabog (il faut desactive OBT)
		// il faudrait quand meme que je verifie que les LR utilisees dans les VLR sont elimines sinon cela baisse beaucoup le N50
		BuildVeryLongReads(&read_booster,verylongreadsfile,&max_mem_usage);
	}
	
	if( read_booster.debug )
	{
		InitMaxMemUsage(&max_mem_usage);
		fprintf(stderr,"Max Memory Usage: %d Mb\n",(int)max_mem_usage/1024);
	}
	
	CleanUpBoostr(&read_booster,CONTIGER_CLEAN_ALL);
	DeleteSeqCount(&h_seqfile2counts,FALSE);
	ParamParserDeleteListOfFiles(&o_param,lof);
	ParamParserDeleteListOfFiles(&o_param,lof_clean);
	ParamParserDelete(&o_param);
	Date("END");
	exit(0);
}

/*
 * 
 */


void ParseReads (read_booster_t *read_booster,lof_t lof,seqcount_t **seq2counts, short mode,short depth)
{
	pid_t pid = getpid();
	unsigned long nb_seq;
	int i=0;
	
	read_booster->process_done=FALSE;
	
	int j;
	for(j=0;j<read_booster->nb_cpus;j++)
		read_booster->current_nt_size_per_threads[j]=0;
	
	/*
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		read_booster->nextid_per_threads_curr[i]=0;
		read_booster->nextid_per_threads_max[i]=0;
	}
	*/
	
	while(lof[i] != NULL && ! read_booster->process_done )
	{
		nb_seq = GetNumberOfSequencePairs(seq2counts,lof[i],read_booster->paired_files,FALSE,FALSE);
		fprintf(stderr,"%ld sequence pairs in file %s\n",nb_seq,lof[i]);

		int step=GetFormat(lof[i]);
		
		arg_threads_t argt;
		argt.complex_arg[0]=read_booster;
		argt.complex_arg[1]=lof[i];
		if ( read_booster->paired_files )
			argt.complex_arg[2]=lof[i+1];
		else
			argt.complex_arg[2]=NULL;
		argt.complex_arg[3]=&mode;
		argt.complex_arg[4]=&depth;
		argt.complex_arg[5]=&pid;
		argt.complex_arg[6]=&step;
		
		InitSMPStatusBar(read_booster->nb_cpus,&argt);
		//MultithreadingNoSlice(ParseFileOfReads,&argt,nb_seq,read_booster->nb_cpus);
		// je reviens au //isme precedent je pense que je cree trop de concurrence car les clusters proches sont traites en meme temps
		Multithreading(ParseFileOfReads,&argt,nb_seq,read_booster->nb_cpus);
		CloseSMPStatusBar(read_booster->nb_cpus,&argt);
		
	 	i += ( read_booster->paired_files ) ? 2 : 1;
	}
	
	// a la fin du parsing cumule le current + le total par thread
	SetCurrentNtSize(read_booster);
	/*
	for(j=0;j<read_booster->nb_cpus;j++)
		if ( read_booster->nextid_per_threads_curr[j] > read_booster->nextid )
			read_booster->nextid = read_booster->nextid_per_threads_curr[j];
	*/
	fprintf(stderr,"==========\nID=%u\n",read_booster->nextid);
		
	if (read_booster->process_done)
		fprintf(stderr,"Parsing stopped, maximum coverage reached ( %llu > ( %llu * %d ) )\n",read_booster->current_nt_size,read_booster->expected_nt_size,depth);
}

void *ParseFileOfReads(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	//fprintf(stderr,"imin=%ld, imax=%ld, thread_id=%d\n",thr_data->imin, thr_data->imax,thr_data->thread_id);
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	char   *file1=(char *)thr_data->complex_arg[1];
	char   *file2=(char *)thr_data->complex_arg[2];
	short  mode=*((short *)thr_data->complex_arg[3]);
	short  depth=*((short *)thr_data->complex_arg[4]);
	pid_t pid=*((pid_t *)thr_data->complex_arg[5]);
	int    step=*((int *)thr_data->complex_arg[6]);
	
	unsigned long long decime=(imax-imin)/10;
	if ( decime == 0 )
		decime=1;
	
	long long k=-1; // je ne peux pas utiliser unsigned
	char read1[MAXREADNT]="";
	char read2[MAXREADNT]="";
	short used   = 0;
	unsigned long long cpt_used = 0;
	int done     = 0;
	
	FILE *f_reads  = GetStreamIn(file1);
	FILE *f_reads2 = NULL;
	if ( file2 != NULL )
		f_reads2 = GetStreamIn(file2);
	char outfile[5000]="";
	char basename[5000]="";
	if (rindex(file1,'/') != NULL )
		strcpy(basename,rindex(file1,'/')+1);
	else
		strcpy(basename,file1);
	sprintf(outfile,"%s/boost-r.log.%02d.%d.%llu-%llu.%d.%s.log",read_booster->outdir,(int)thr_data->thread_id,(int)pid,imin,imax,(int)mode,basename);
//	fprintf(stderr,"LOG: %s\n",outfile);
//	fflush(stderr);
	FILE *f_log=NULL;
	if ( read_booster->debug )
		f_log=GetStreamOut(outfile);
	unsigned long long cpt = 0;
	int j;
	char line[MAXLINE];
	int readdone=FALSE;
	int lenr1,lenr2;
	unsigned long long nt_used_in_thread = 0;
	while((fgets(line,MAXLINE,f_reads) != NULL) && ! readdone)
	{
		k++;
		cpt++;
		if ( k > imax )
		{
			readdone=TRUE;
			continue;
		}
		for(j=1;j<step*2;j++)
		{
			if ( read_booster->paired_files && j >= step )
			{
				if ( fgets(line,MAXLINE,f_reads2) == NULL )
					exit(EXIT_FAILURE);
			}
			else if ( fgets(line,MAXLINE,f_reads) == NULL )
				exit(EXIT_FAILURE);
			if ( k >= imin && k <= imax )
			{
				if (j==1)
				{
					//strncpy(read1,line,read_booster->max_readlen_used);
					//read1[read_booster->max_readlen_used]='\0';
					strcpy(read1,line);
					lenr1=strlen(line);
					read1[lenr1-1]='\0';  // retour chariot
				}
				else if ( j == (1+step) )
				{
					//strncpy(read2,line,read_booster->max_readlen_used);
					//read2[read_booster->max_readlen_used]='\0';
					strcpy(read2,line);
					lenr2=strlen(line);
					read2[lenr2-1]='\0';  // retour chariot
				}
			}
		}
//		if ( k % nb_cpus == thr_data->thread_id )
		if ( k >= imin && k <= imax )
		{
			//fprintf(stderr,"\t%d = %lld %llu-%llu, %llu, %llu\n",thr_data->thread_id,k,imin,imax,k-imin,imax-imin+1);
			if ( (k-imin) % decime == 0 )
				SMPStatusBar(thr_data->thread_id,(unsigned long long)((unsigned long long)k-imin),(unsigned long long)(imax-imin+1),thr_data);
			// 20140329, cf step 0 if ( index(read1,'N') == NULL && index(read2,'N') == NULL && strlen(read1) >= read_booster->min_readlen_foruse && strlen(read2) >= read_booster->min_readlen_foruse )
			// {
			// je ne peux pas trop deplacer IsKmerCompositionOK dans step 0 car je crois que je modifie le seuil dans fx : A VERIFIER
			// c'est bien important de le faire sur le compte des kmer final (apres le cln et apres le resize eventuel)
			//	if ( read_booster->a_kmercode == NULL || ( IsKmerCompositionOK(read_booster,read1,TRUE) && IsKmerCompositionOK(read_booster,read2,TRUE) ) ) // pas de klib ou compo ok
			//	{
					used=0;
					//float qr1,qr2;
					switch(mode)
					{
						case ONLY_UNMAPPED:
							AddPairOfReads(read_booster,read1,read2,ONLY_UNMAPPED,depth,&used,thr_data->thread_id);
							break;
						case INCREASE_DEPTH:
							AddPairOfReads(read_booster,read1,read2,INCREASE_DEPTH,depth,&used,thr_data->thread_id);
							break;
						case REPLACE_PROBLEMATIC_PAIRS:
							ReplaceProblematicPairs(read_booster,read1,read2,depth,&used);
							break;
						case RESCUE_MISSING_KMERS:
							RescuePairsWithFrequentsKmers(read_booster,read1,read2,depth,&used);
							break;
						// pour HT-BES
						case ONLY_UNMAPPED_PAIRS:
							AddPairOfReads(read_booster,read1,read2,ONLY_UNMAPPED_PAIRS,depth,&used,thr_data->thread_id);
							break;
						case INCREASE_DEPTH_PAIRS:
							AddPairOfReads(read_booster,read1,read2,INCREASE_DEPTH_PAIRS,depth,&used,thr_data->thread_id);
							break;
						default:
							fprintf(stderr,"Error: unmanaged mode %d\n",mode);
							break;
					}
					done=FALSE;
					if ( used )
					{
						nt_used_in_thread+=(lenr1+lenr2);
						cpt_used++;
					}
					if ( cpt_used && (k % read_booster->step_thread) == 0 && mode != RESCUE_MISSING_KMERS) // je ne teste plus a chaque lecture au niveau de chaque thread mais au niveau global (en moyenne)
					{
						if ( mode ==  REPLACE_PROBLEMATIC_PAIRS )
							done = TestMaxCoverage(read_booster,depth);
						else 
						{
							// pour eviter la synchro lorsqu'il y a des collectes par threads
							// je ne commence a faire le test que si l'extrapolation du nombre de lectures conservees depasse 3/4 de la profondeur visee
							unsigned long long nt_testing_limit = ((unsigned long long)read_booster->depth * 3 * (unsigned long long)read_booster->expected_nt_size / 4);
							// estimation de la taille globale si tous les threads ont conserves le meme nombre de reads
							// ce sont les threads les plus en avance qui vont sans doute declencher l'arret
							if ( (nt_used_in_thread*(unsigned long long)read_booster->nb_cpus) >= nt_testing_limit)
							{
//								fprintf(stderr,"Test depth: %llu %llu %llu\n",nt_used_in_thread*(unsigned long long)read_booster->nb_cpus,nt_testing_limit,(unsigned long long)read_booster->depth * (unsigned long long)read_booster->expected_nt_size);
								done = TestMaxCoverage(read_booster,depth);
							}
						}
					}
					if ( read_booster->process_done || done )
					{
						if ( done )
						{
							if (pthread_rwlock_wrlock(&read_booster->lock_process_done) != 0) Fatal("can't get wrlock");
							read_booster->process_done=TRUE;
							pthread_rwlock_unlock(&read_booster->lock_process_done);
						}
						if ( f_log != NULL )
							fclose(f_log);
						CloseStream(file1,f_reads);
						if ( file2 != NULL )
							CloseStream(file2,f_reads2);
						return(NULL);
					}
			//	}
			//}
			if ( cpt % 100000 == 0 && cpt && f_log != NULL ) 
			{
				float ratio = (float)cpt_used / 100000;
				fprintf(f_log,"read:%llu used:%llu ratio=%.4f cov=%llu\n",k+1,cpt_used,ratio,GetCurrentCoverage(read_booster));
				fflush(f_log);
				cpt_used = 0;
			}
		}
	}
	CloseStream(file1,f_reads);
	if ( file2 != NULL )
		CloseStream(file2,f_reads2);
	if ( f_log != NULL )
		fclose(f_log);
	return(NULL);
}



void Usage (char *argv[])
{
	fprintf(stderr,"\
\n\
Boost your reads - boost-r, cast from short reads to long reads\n\
Usage: boost-r --short '../data/*.fastq.gz' --size 5 --long longreads\n\
\n\
Mandatory\n\
--short        [quoted]string        list of paired/interleaved fast[aq] files (ex: '../data/*.fastq.gz')\n\
                                     warning: do not forget the single quotes '' for a list of files\n\
--long             string            [boost-r] prefix of the output files\n\
--size             float             expected/estimated genome size (in Mb)\n\
\n\
Optional\n\
--klen             integer           [%d] length of kmers\n\
--depth            integer           [%d] max depth\n\
--nb_threads       integer           [%d] number of threads (Warning: the results depend on this value)\n\
--interleaved                        flag to be used when read1 and read2 are interleaved in the same files\n\
--insert_size      integer           [%d] set the mean insert size in raw data (the max if several libs are used)\n\
--outdir           string            [.] directory where results will be saved\n\
--illumina_adapt   string            illumina adapter sequences (quoted string, space separator)\n\
--compress         integer           [1] set to 0 to deactivate file compression (faster)\n\
--maxpaircov       integer           [%d] maximum number of links between read pairs (control redundancy)\n\
Shortcuts\n\
--preprocessed_LR  string            multifasta file of long reads\n\
--preprocessed_reads string          multifasta file of cleaned reads\n\
Alignement cut-offs\n\
--minovl           integer           [%d] minimum overlap length (nt)\n\
--minpci           float             [%.2f] minimum percentage identity of the overlap\n\
--maxovh           integer           [%d] maximum length of the overhang (nt)\n\
--maxmm            integer           [%d] maximum mismatch number\n\
--minkmers_for_lr_extension integer  minimum number of shared kmers required for running alignement\n\
                                     default value depends on klen, (klen < 14) ? 3 : 2 \n\
--minkdepth        integer           [%d] minimum kmer occurence\n\
Debug\n\
--build_long_reads integer           [1] set to 0 to deactivate the computation of long reads (LR)\n\
--build_very_long_reads integer      [1] set to 0 to deactivate the computation of very long reads (VLR)\n\
--dump_klib        string            dump the kmer library when defined\n\
--load_klib        string            load the kmer library from the given file\n\
--save_reads                         dump the read pairs after each filtering step\n\
--steps            string            [01234] list of steps for read selection\n"
,DEFAULT_PARAM_KLEN,DEFAULT_PARAM_DEPTH,DEFAULT_PARAM_NB_THREADS,DEFAULT_PARAM_INSERT_SIZE,DEFAULT_PARAM_MAXPAIRCOV,DEFAULT_PARAM_MINOVL,(float)DEFAULT_PARAM_MINPCI,DEFAULT_PARAM_MAXOVH,DEFAULT_PARAM_MAXMM,DEFAULT_PARAM_MINKDEPTH);
}
