/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: general.h 517 2016-08-17 14:31:58Z gouzy $
 * 
 ---------------------------------*/

#ifndef GENERAL_H

#define GENERAL_H

#include "pthread.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "limits.h"
#include <openssl/md5.h>
#include "assert.h"
#include "time.h"
#include "math.h"
#include <unistd.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <ifaddrs.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/resource.h>

#if defined(__MIC__)
	#include <mm_malloc.h>
#endif

#define COMPRESSED_NO    0
#define COMPRESSED_GZIP  1
#define COMPRESSED_BZIP2 2
#define COMPRESSED_XZ    3

/* attention aux effets de bord pour Min/Max, ne pas imbriquer d'appel a des fonctions */
#define Min(X, Y)  ((X) < (Y) ? (X) : (Y))
#define Max(X, Y)  ((X) > (Y) ? (X) : (Y))

#define TRUE 1
#define FALSE 0
#define MAXLINE 	1000000
#define MAXFILENAME 10000
#define MAX_CPUS             1024
#define MULTITHREADING_ERROR 1
#define PTHREAD_MAX_COMPLEX_ARGS 100

typedef struct arg_threads
{
	unsigned long long imin;
	unsigned long long imax;
	int ncpus;
	int job_id;
	int  thread_id;
	short running;
	pthread_rwlock_t   *ptr_lock_self_array;
	struct arg_threads *self_array;
	pthread_rwlock_t   *ptr_lock;
	char *statusbar;
	void *complex_arg[PTHREAD_MAX_COMPLEX_ARGS];
} arg_threads_t;

void  Fatal(char *msg);
void  Date(char *tag);
int   GetExtension(char *file);
FILE *GetStreamIn(char *file);
void  CloseStream(char *file,FILE *f_in);
FILE *GetStreamOut(char *file);
int   GetListOfFiles(char *pattern,char *lof);
int   Split(char *str,int c,char *a_split[]);
int   StrEndWith(char *str,char *substr);
int   StrStartWith(char *str,char *substr);
char *GetTime(void);
unsigned long long StatusBar(unsigned long long done,unsigned long long max);
void  CloseStatusBar(void);
unsigned long long SMPStatusBar(short cpu_id,unsigned long long done, unsigned long long max,arg_threads_t *argt);
void  CloseSMPStatusBar(int nb_cpus,arg_threads_t *argt);
void  InitSMPStatusBar(int nb_cpus,arg_threads_t *argt);
void CheckIPAuth(char lstok[]);
char *GetListOfIP(void);
void InitMaxMemUsage(unsigned long long *max_mem);
void FreeSplit(char **a_strs,int nb);
void SetStackSize(int kbvalue);
char *StrRStr(char *str,char *substr);


int Multithreading(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus);
int MultithreadingNoSlice(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus);
int MultithreadingShareSliceWithHelper(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long imin, unsigned long long imax, int currthread_id,int helperthread_id,int job_id);
int MultithreadingNoSliceReuse(void *(*OneThread)(void*),arg_threads_t *argt,unsigned long long maxjobs,int nb_cpus,arg_threads_t *a_argt);
void FreeArrayOfThreadArgsIfNeeded(arg_threads_t *argt,arg_threads_t *a_argt);

int GetFinishedThreadId(arg_threads_t *argt);
void SetThreadIdAsFinished(arg_threads_t *argt,int thread_id);
void SetThreadIdAsRunning(arg_threads_t *argt,int thread_id);

void *AlignedMallocOnPageSize(size_t size);


#endif