/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140407
 * Copyright INRA, All rights reserved
 * 
 * $Id: rd_utils.c 491 2016-03-17 15:19:44Z gouzy $
 * 
 * ---------------------------------*/

#include "rd_utils.h"

// phymap: si min_rdlen_for_countnt je compte le nombre de nt dans des reads > min_rdlen_for_countnt et je cree une cle filename-ntcount dans h_seqcount
unsigned long long GetNumberOfSequencePairs (seqcount_t **h_seqcount,char *file,int paired_files,int single_end,int min_rdlen_for_countnt)
{
	
	// fprintf(stderr,"========== %s ============ paired_files=%d, single_end=%d\n",file,paired_files,single_end);
	seqcount_t *oneseqcount=NULL;
	seqcount_t *oneseqntcount=NULL;
	
	assert(file);
		
	HASH_FIND_STR(*h_seqcount, file, oneseqcount);
	if ( oneseqcount )
	{
	//	fprintf(stderr,"GetNumberOfSequencePairs (%s) precomputed %lld\n",file,oneseqcount->count);
		return(oneseqcount->count);
	}
	
	unsigned long long cpt=0;
	unsigned long long j;
	char line[MAXLINE];
	
	int step=GetFormat(file);

	FILE *f_pairs = GetStreamIn(file);
	int fact = ( paired_files ) ? 1 : 2;
	if ( single_end )
	{
		paired_files = FALSE;
		fact = 1;
	}
	unsigned long long ntcount=0;
	int limit=step*fact;
	int len;
//	fprintf(stderr,"step(%d)*fact(%d)=%d\n",step,fact,step*fact);
	while(fgets(line,MAXLINE,f_pairs) != NULL)
	{
		cpt++;
		for(j=1;j<limit;j++)
		{
			if ( fgets(line,MAXLINE,f_pairs) == NULL )
			{
				fprintf(stderr,"Error: likely corrupted file %s\n",file);
				if ( (double)(((cpt-1)*step*fact+j) % (step*fact)) != 0 )
					fprintf(stderr,"The number of lines of this file is expected to be modulo %d\n",step*fact);
				exit(EXIT_FAILURE);
			}
			len=strlen(line);
			if ( j == 1 && min_rdlen_for_countnt && len >= min_rdlen_for_countnt ) // je ne veux compter que ce qui depasse un seuil
				ntcount+=len;
		}
	}
	CloseStream(file,f_pairs);
    oneseqcount=(seqcount_t *)malloc(sizeof(seqcount_t));
	assert(oneseqcount);
	oneseqcount->count=cpt;
	len = strlen(file);
	oneseqcount->filename=(char *)malloc(len+1);
	assert(oneseqcount->filename);
	strcpy(oneseqcount->filename,file);
	HASH_ADD_KEYPTR(hh,*h_seqcount,oneseqcount->filename,len,oneseqcount);
	if ( ! paired_files && ! single_end )
		cpt *= 2;
	if (min_rdlen_for_countnt)
	{
	    oneseqntcount=(seqcount_t *)malloc(sizeof(seqcount_t));
		assert(oneseqntcount);
		oneseqntcount->count=ntcount;
		int len = strlen(file)+strlen("-ntcount");
		oneseqntcount->filename=(char *)malloc(len+1);
		assert(oneseqntcount->filename);
		sprintf(oneseqntcount->filename,"%s-ntcount",file);
		HASH_ADD_KEYPTR(hh,*h_seqcount,oneseqntcount->filename,len,oneseqntcount);
	}
	return cpt;
}

int GetFormat(char file[])
{
	FILE *f_reads = GetStreamIn(file);
	char c =  fgetc(f_reads);
	int format;
	switch(c)
	{
		case '>':
			format=2;
			break;
		case '@':
			format=4;
			break;
		default:
			fprintf(stderr,"Error: unknown file format or empty file (%s)\n",file);
			exit(EXIT_FAILURE);
	}
	CloseStream(file,f_reads);
	return(format);
}

void ResetKlib(read_booster_t *read_booster)
{
	short maxk_for_alloc = Min(read_booster->klen,16);
	unsigned long long max=pow(4,maxk_for_alloc);
	unsigned long long idx;
	for(idx=0;idx<max;idx++)
		read_booster->a_kmercode[idx]=0;
}

void ResetReadFileAndKlib(read_booster_t *read_booster,paramparser_t **o_param,char *file,lof_t lof,int initklib,seqcount_t **h_seqfile2counts,int revcomp)
{
	ParamParserSet(o_param,"short",file);
	read_booster->paired_files=FALSE;
	ParamParserDeleteListOfFiles(o_param,lof);
	ParamParserGetListOfFiles(o_param,"short",lof);
	AssertDataSets(read_booster,lof);
	if ( initklib )
	{
		if ( read_booster->debug )
			Date("UPDATING KLIB WITH SELECTED PAIRS");
		if ( read_booster->a_kmercode != NULL )
			ResetKlib(read_booster);
		BuildKmerLibrary(read_booster,lof,h_seqfile2counts,revcomp,FALSE,FALSE);
	}
}

/*
 * 
 */

void AssertDataSets(read_booster_t *read_booster, lof_t lof)
{
	int i;
	for(i=0;lof[i]!=NULL;i++)
	{
		fprintf(stderr,"Data file(s): %s\n",lof[i]);
	}
	if ( read_booster->paired_files && i % 2 != 0 )
	{
		fprintf(stderr,"Error: paired_files set, an even number of files is required\n");
		exit(EXIT_FAILURE);
	}
}

/*
 * 
 */



void BuildKmerLibraryFromSelection (read_booster_t *read_booster,int revcomp)
{
	Date("RVM");
	int klen=read_booster->klen;
	if ( read_booster->a_kmercode_selection != NULL )
		free(read_booster->a_kmercode_selection);
	short maxk_for_alloc = Min(klen,16);
	read_booster->a_kmercode_selection=(short *)calloc(pow(4,maxk_for_alloc),sizeof(kmercode_t));
	
	arg_threads_t argt;
	argt.complex_arg[0]=read_booster;
	argt.complex_arg[1]=&revcomp;
	unsigned long long nb_seq = read_booster->nextid;
	
	InitSMPStatusBar(read_booster->nb_cpus,&argt);
	Multithreading(BuildKmerLibrarySelectionSlice,&argt,nb_seq,read_booster->nb_cpus);
	CloseSMPStatusBar(read_booster->nb_cpus,&argt);
}

void *BuildKmerLibrarySelectionSlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	int    revcomp=*((int *)thr_data->complex_arg[1]);
	
	unsigned long long decime=(imax-imin)/10;
	if ( decime == 0 )
		decime=1;
	
	int klen = read_booster->klen;
	unsigned long long i,k;
	int lmax;
	char *ptr_str;
	char *rcseq;
	kmercode_t kmercode;
	pthread_rwlock_t *ptr_lock;
	
	for(i=imin;i<=imax;i++)
	{
		char *seq = GetSeq(read_booster->h_seqcontigs,i);
		if (seq == NULL )
			continue;
		
		if ( (i-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(i-imin),(unsigned long long)(imax-imin+1),thr_data);
		
		int len=strlen(seq);
		lmax = len-klen;
		ptr_str=seq;
		short init = FALSE;
		for(k=0;k<=lmax;k++,ptr_str++)
		{	
			init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
			if ( ! init )
				continue;
			if ( read_booster->a_kmercode_selection[kmercode] == SHRT_MAX )
				continue;
			
			ptr_lock=&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS];
			if (pthread_rwlock_wrlock(ptr_lock) != 0) Fatal("can't get wrlock");
			read_booster->a_kmercode_selection[kmercode]++;
			pthread_rwlock_unlock(ptr_lock);
		}
		if ( revcomp )
		{
			rcseq=RevCompSeq(seq);
			ptr_str=rcseq;
			short init = FALSE;
			for(k=0;k<=lmax;k++,ptr_str++)
			{	
				init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
				if ( ! init )
					continue;
				if ( read_booster->a_kmercode_selection[kmercode] == SHRT_MAX )
					continue;
				
				ptr_lock=&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS];
				if (pthread_rwlock_wrlock(ptr_lock) != 0) Fatal("can't get wrlock");
				read_booster->a_kmercode_selection[kmercode]++;
				pthread_rwlock_unlock(ptr_lock);
			}
			free(rcseq);
		}
	}
	return(NULL);
}
/*
 * 
 */

void BuildKmerLibrary (read_booster_t *read_booster,lof_t lof,seqcount_t **seq2counts,int revcomp,int update,int ntcount)
{
	int done=FALSE;
	unsigned long long nb_seq;
	int i=0;
	int step;
	int klen=read_booster->klen;
	short maxk_for_alloc = Min(klen,16);
	if ( ! update )
	{
		if ( read_booster->a_kmercode != NULL )
			free(read_booster->a_kmercode);
		read_booster->a_kmercode=(short *)calloc(pow(4,maxk_for_alloc),sizeof(kmercode_t));
		assert(read_booster->a_kmercode != NULL);
	}
	//ResetKlib(read_booster);
	while(lof[i] != NULL && ! done )
	{
		nb_seq = GetNumberOfSequencePairs(seq2counts,lof[i],read_booster->paired_files,read_booster->single_end,ntcount);
		fprintf(stderr,"%llu reads in %s\n",nb_seq,lof[i]);
		step=GetFormat(lof[i]);
		Date("KLIB intra");
		
		arg_threads_t argt;
		argt.complex_arg[0]=read_booster;
		argt.complex_arg[1]=lof[i];
		argt.complex_arg[2]=&step;
		argt.complex_arg[3]=&revcomp;
		argt.complex_arg[4]=&update;
		
		InitSMPStatusBar(read_booster->nb_cpus,&argt);
		MultithreadingNoSlice(BuildKmerLibrarySlice,&argt,nb_seq,read_booster->nb_cpus);
		CloseSMPStatusBar(read_booster->nb_cpus,&argt);
		i++;
	}
}	

void *BuildKmerLibrarySlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	char   *file=(char *)thr_data->complex_arg[1];
	unsigned long long    step=*((int *)thr_data->complex_arg[2]);
	int    revcomp=*((int *)thr_data->complex_arg[3]);
	int    update=*((int *)thr_data->complex_arg[4]);
	
	unsigned long long decime=(imax-imin)/10;
	if ( decime == 0 )
		decime=1;
	
	int nb_cpus = read_booster->nb_cpus;
	
	int maxrdlen=read_booster->min_readlen_foruse;
	short klen=read_booster->klen;
	int lmax=maxrdlen-klen;
	FILE *f_reads = GetStreamIn(file);
	char line[MAXLINE];
	char read[MAXREADNT]="";
	int k,j;
	kmercode_t kmercode;
	char *ptr_str;
	unsigned long long cpt=0;
	int done = FALSE;
	char  *sseq;
	pthread_rwlock_t *ptr_lock;
	int len = 0;
	char *rcseq=NULL;
	
	while((fgets(line,MAXLINE,f_reads) != NULL) && ! done)
	{
		cpt++;
		if ( 0 && cpt > imax )
		{
			done=TRUE;
			continue;
		}
		for(j=1;j<step;j++)
		{
			if ( fgets(line,MAXLINE,f_reads) == NULL )
				exit(EXIT_FAILURE);
			if ( j == 1 )
			{
				/* 20140526 je change au moment du probleme sur les nok == ATTENTION
				strncpy(read,line,maxrdlen);
				read[maxrdlen]='\0';
				if ( (sseq=index(read,'N')) != NULL )
					*sseq='\0';
				*/
				// 20151228: passage en memcpy, ca n'arrange rien c'est la lecture qui bloque
				len=strlen(line)-1; // retour chariot
				memcpy(read,line,len);
				read[len]='\0';
				
				// 20151228: meme si je supprime (pas bon pour illumina) ca n'arrange rien c'est la lecture qui bloque
				if ( (sseq=index(read,'N')) != NULL )
				{
					*sseq='\0';
					len=strlen(read);
				}
			}
		}
		if ( (cpt-imin) % decime == 0 )
			SMPStatusBar(thr_data->thread_id,(unsigned long long)(cpt-imin),(unsigned long long)(imax-imin+1),thr_data);
		
//		if ( cpt >= imin && cpt <= imax )
		if ( cpt % nb_cpus == thr_data->thread_id )
		{
			if ( len >= maxrdlen )
			{
				lmax = len-klen;
				ptr_str=&read[0];
				short init = FALSE;
				//lmax=len-klen; // 20140526: je change au moment du probleme de perf sur les nok, avant c'etait toujours maxrdlen-klen
				for(k=0;k<=lmax;k++,ptr_str++)
				{	
					init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
					if ( ! init )
						continue;
					if ( read_booster->a_kmercode[kmercode] == SHRT_MAX )
					{
						/*
						if ( update )
						{
							k+=(klen-1);
							ptr_str+=(klen-1);
							init=FALSE;
						}
						*/
						continue;
					}
					
					ptr_lock=&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS];
					if (pthread_rwlock_wrlock(ptr_lock) != 0) Fatal("can't get wrlock");
					read_booster->a_kmercode[kmercode]++;
					pthread_rwlock_unlock(ptr_lock);
				}
				if ( revcomp )
				{
					rcseq=RevCompSeq(read);
					ptr_str=rcseq;
					short init = FALSE;
					for(k=0;k<=lmax;k++,ptr_str++)
					{	
						init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
						if ( ! init )
							continue;
						if ( read_booster->a_kmercode[kmercode] == SHRT_MAX )
						{
							/*
							if ( update )
							{
								k+=(klen-1);
								ptr_str+=(klen-1);
								init=FALSE;
							}
							*/
							continue;
						}
						ptr_lock=&read_booster->a_lock_a_kmer2contigs[kmercode%MAX_THREAD_LOCKS];
						if (pthread_rwlock_wrlock(ptr_lock) != 0) Fatal("can't get wrlock");
						read_booster->a_kmercode[kmercode]++;
						pthread_rwlock_unlock(ptr_lock);
					}
					free(rcseq);
				}
			}
		}
	}
	CloseStream(file,f_reads);
	return(NULL);
}

void SaveKlib(read_booster_t *read_booster,char filename[])
{
	FILE *f_out = GetStreamOut(filename);
	unsigned long long idx;
	unsigned long long max;
	int klen=read_booster->klen;
	if ( klen > 16 )
		Fatal("Error: SaveKlib fonction cannot be used with klen > 16\n");
	short maxk_for_alloc=Min(klen,16);
	max=pow(4,maxk_for_alloc);
	char kmerdec[255];
	// je ne fais pas le revcomp, chaque kmer doit etre vu sur les deux brins
	for(idx=0;idx<max;idx++)
	{
		if ( read_booster->a_kmercode[idx] >= read_booster->minkdepth )
		{
			KmerDecode(idx,klen,kmerdec);
			fprintf(f_out,"%s\t%d\t%llu\n",kmerdec,read_booster->a_kmercode[idx],idx);
		}
	}
	CloseStream(filename,f_out);
}


/*
 * Prerequesite: la klib ne doit pas etre initialisee
 */

// Format: kmer count kmercode
void LoadKlib(read_booster_t *read_booster,char filename[])
{
	kmercode_t kmercode;
	int klen=read_booster->klen;
	short maxk_for_alloc = Min(klen,16);
	read_booster->a_kmercode=(short *)calloc(pow(4,maxk_for_alloc),sizeof(kmercode_t));
	assert(read_booster->a_kmercode);
	FILE *f_in = GetStreamIn(filename);
	char kmer[255];
	char line[MAXLINE];
	int occur;
	while(fgets(line,MAXLINE,f_in) != NULL)
	{
		sscanf(line,"%s %d %u",kmer,&occur,&kmercode);
		read_booster->a_kmercode[kmercode]=(short)occur;
	}
	CloseStream(filename,f_in);
}

// Format kmer count
void LoadKlib2(read_booster_t *read_booster,char filename[])
{
	kmercode_t kmercode;
	int klen=read_booster->klen;
	short maxk_for_alloc = Min(klen,16);
	read_booster->a_kmercode=(short *)calloc(pow(4,maxk_for_alloc),sizeof(kmercode_t));
	assert(read_booster->a_kmercode);
	FILE *f_in = GetStreamIn(filename);
	char kmer[255];
	char line[MAXLINE];
	int occur;
	while(fgets(line,MAXLINE,f_in) != NULL)
	{
		sscanf(line,"%s %d",kmer,&occur);
		KmerEncode(kmer,klen,&kmercode);
		read_booster->a_kmercode[kmercode]=occur;
	}
	CloseStream(filename,f_in);
}


// verifier pour les klen eleves que cela marche toujours
int IsAdapterPresent(read_booster_t *read_booster,char *read)
{
	int klen=read_booster->klen;
	int lmax=strlen(read)-klen;
	char *ptr_str=read;
	int i;
	kmercode_t kmercode;
	short init = FALSE;
#ifndef NDEBUG
	assert((lmax+klen-1)<strlen(read));
#endif
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( i == lmax || ! init || read_booster->a_kmercode[kmercode] < read_booster->minkdepth )
		{
			if ( (i+klen) >= ADAPT_MINLEN )
				return(TRUE);
			else
				return(FALSE);
		}
	}
	return(FALSE);
}

int IsKmerCompositionOK(read_booster_t *read_booster,char *read,short last)
{
	int klen=read_booster->klen;
	int lmax=(read_booster->maxkmers-1)*klen;
	if ( read_booster->offsetk > 0 )
		lmax-=(read_booster->offsetk*(read_booster->maxkmers-1));
#ifndef NDEBUG
	assert((lmax+klen)<=strlen(read));
#endif
	char *ptr_str=read;
	int i;
	kmercode_t kmercode;
	short init = FALSE;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( ! init )
			return(FALSE);
		if ( read_booster->a_kmercode[kmercode] < read_booster->minkdepth )
			return(FALSE);
	}
	if (last )
	{
		// 20140515
		// j'ai verifie sur la longeur utile minimum mais j'essaye de prendre plus si possible
		// 20140515: si le dernier kmer possible est ok je prend toute la lecture sinon je coupe 
		short lastkpos=strlen(read)-klen;
		init = KmerEncode(read+lastkpos,klen,&kmercode);
//		fprintf(stderr,"i=%d %d lastkpos:%d %u %d %s\n",i,strlen(read),lastkpos,kmercode,init,read+lastkpos);
		if ( init && read_booster->a_kmercode[kmercode] < (read_booster->depth/3) ) 
			read[lmax+klen]='\0';
	}
	return(TRUE);
}

int TrimReadWithRareKmers(read_booster_t *read_booster,char *read)
{
	int klen=read_booster->klen;
	int readlen=strlen(read);
	int lmax=readlen-klen;
	char *ptr_str=read;
	int maxi=(read_booster->maxkmers-1)*klen;
	if ( read_booster->offsetk > 0 )
		maxi-=(read_booster->offsetk*(read_booster->maxkmers-1));
#ifndef NDEBUG
	assert((lmax+klen-1)<strlen(read));
	assert((maxi+klen-1)<strlen(read));
#endif
	int i;
	kmercode_t kmercode;
	short init = FALSE;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		init = KmerEncodeHub(ptr_str,klen,&kmercode,read_booster->kmercode_mask,init);
		if ( ! init || read_booster->a_kmercode[kmercode] < read_booster->minkdepth )
		{
			if ( i > maxi ) // il faut que je sois ok si j'ai une longeur suffisante de ok correspondant a min_readlen_foruse (je mets read_booster->maxkmers*klen pour essayer de traiter des reads plus longs)
				i=(maxi+klen);
			read[i]='\0';
			return(i); // longeur de la nouvelle sequence
		}
	}
	return(readlen);
}


// je l'utilise pour phix

int RemoveAdapterByKmer(read_booster_t *read_booster,char *read,dejavucode_t **h_phix)
{
	int klen=15; // pour eviter des soucis de sensibilites sur les kmers eleves
	int readlen=strlen(read);
	int lmax=readlen-klen;
	char *ptr_str=read;
	int i;
	kmercode_t kmercode;
	kmercode_t kmercode_mask = pow(4,15-1)-1;
	short init = FALSE;
	int found = FALSE;
	dejavucode_t *deja=NULL;
	for(i=0;i<=lmax;i++,ptr_str++)
	{	
		if ( ! init )
			init = KmerEncode(ptr_str,klen,&kmercode);
		else
			init = KmerEncodeOS(ptr_str,klen,&kmercode,kmercode_mask);
		if ( init )
		{
			// j'ai mis aussi le recomp de phix dans le hash
			HASH_FIND(hh, *h_phix, &kmercode, sizeof(kmercode_t), deja);
			if ( deja != NULL )
				found++;
		}
	}
	return(found > (lmax/3));
}


void HashAdapter(read_booster_t *read_booster,char *seq,dejavucode_t **h_phix)
{
	int klen=15; // je ne veux pas que cela depende du klen general
	int i,j;
	char* a_seqs[50]; // 50 sequences d'adapters max
	for (i=0;i<50;i++)
		a_seqs[i]=NULL;
	Split(seq,' ',a_seqs);
	kmercode_t kmercode;
	kmercode_t kmercode_mask = pow(4,15-1)-1;
	int cpt=0;
	j=0;
	while(a_seqs[j]!=NULL)
	{
//		fprintf(stderr,"%d %s\n",j,a_seqs[j]);
		int lmax = strlen(a_seqs[j])-klen;
		int init=FALSE;
		char *ptr_str=seq;
		for(i=0;i<=lmax;i++,ptr_str++)
		{	
			if ( ! init )
				init = KmerEncode(ptr_str,klen,&kmercode);
			else
				init = KmerEncodeOS(ptr_str,klen,&kmercode,kmercode_mask);
			if ( init )
			{
				AddIfNotCodeDejaVu(h_phix,kmercode,klen,FALSE);
				AddIfNotCodeDejaVu(h_phix,kmercode,klen,TRUE);
			}
		}
		cpt++;
		j++;
	}
	FreeSplit(a_seqs,cpt);
}

int RemoveAdapterByAlign(char *read,char *a_seqs[],char c,short **p_mat)
{
	int i=0;
	hsp_t hsp;
	int found=FALSE;
	while(!found && a_seqs[i] != NULL)
	{
		PairwiseMatchUngapped(read,a_seqs[i],&hsp,p_mat);
//		PrintHsp(&hsp);
		if ( hsp.len >= 10 && (hsp.mm <= 1 || hsp.pci >= 95) ) // pour les petits hits cela sera mm et pour les grand pci
		{
			if ( hsp.len >= 20 || ( hsp.qend >= (strlen(read)-5) && hsp.tstart <= 5 ) ) // soit le hit est important soit il est pres du bord et ancre sur le start de l'adapter
			{
				read[hsp.qstart]='\0';
				return(hsp.qstart); // longueur de la nouvelle sequence
			}
		}
		i++;
	}
	return(strlen(read));
}


int InitAdapterSetFromDataset(read_booster_t *read_booster,char *a_adapterset[])
{
	int i=0;
	int iok=0;
	while(a_adapterset[i]!=NULL)
	{
//fprintf(stderr,"%d,%d\t%s\t%f\n",i,iok,a_adapterset[i],GetQualScore(read_booster,a_adapterset[i]));
		if ( IsAdapterPresent(read_booster,a_adapterset[i]) )
		{
			if ( i > iok )
			{
				if ( a_adapterset[iok] != NULL )
					free(a_adapterset[iok]);
				a_adapterset[iok]=(char *)malloc(sizeof(char)*(strlen(a_adapterset[i])+1));
				strcpy(a_adapterset[iok],a_adapterset[i]);
			}
			iok++;
		}
		else if ( a_adapterset[i] != NULL )
		{
			free(a_adapterset[i]);
			a_adapterset[i]=NULL;
		}
		i++;
	}
	a_adapterset[iok]=NULL;
	i=0;
	fprintf(stderr,"Selected adapter(s) sequences: %d\n",iok);
	while(a_adapterset[i]!=NULL)
	{
		fprintf(stderr,"%s\n",a_adapterset[i]);
		i++;
	}
	fprintf(stderr,"========================\n");
	return(i);
}

// je vais faire le IsKmerCompositionOK sur le jeu de donnees resize et deja nettoye (avec klib update) pour ne pas le faire a chaque step
void CleanUpDatasetMore(read_booster_t *read_booster,char *cleanaupdataset)
{
	char cmd[1000];
	char tmpfile[MAXFILENAME]="";
	char ext[MAXFILENAME]="";
	if ( read_booster->compress )
		strcpy(ext,".gz");
	int ret_code;
	
	sprintf(tmpfile,"%s.tmp%s",cleanaupdataset,ext);
	sprintf(cmd,"\\mv %s %s\n",cleanaupdataset,tmpfile);
	ret_code = system(cmd);
	FILE *f_in=GetStreamIn(tmpfile);
	FILE *f_out=GetStreamOut(cleanaupdataset);
	int step=2; // fasta
	unsigned long long k=0;
	char line[MAXLINE];
	int readdone=FALSE;
	char read1[MAXREADNT]="";
	char read2[MAXREADNT]="";
	int j;
	while((fgets(line,MAXLINE,f_in) != NULL) && ! readdone)
	{
		for(j=1;j<step*2;j++)
		{
			if ( fgets(line,MAXLINE,f_in) == NULL )
				exit(EXIT_FAILURE);
			
			if (j==1)
			{
				strcpy(read1,line);
				read1[strlen(line)-1]='\0';  // retour chariot
			}
			else if ( j == (1+step) )
			{
				strcpy(read2,line);
				read2[strlen(line)-1]='\0';  // retour chariot
			}
		}
		if ( IsKmerCompositionOK(read_booster,read1,TRUE) && IsKmerCompositionOK(read_booster,read2,TRUE) )
		{	
			k++;
			fprintf(f_out,">%llu/1\n%s\n>%llu/2\n%s\n",k,read1,k,read2);
		}
	}
	CloseStream(tmpfile,f_in);
	CloseStream(cleanaupdataset,f_out);
	sprintf(cmd,"\\rm %s\n",tmpfile);
	ret_code = system(cmd);
}
	
void CleanUpDataset(read_booster_t *read_booster,lof_t lof,lof_t lof_clean,char *cleanupdataset,char *adapter,char *phix,seqcount_t **seq2counts)
{
//	fprintf(stderr,"Known adapter sequences: %s\n",adapter);
	int idx=0;
	int j=0;
	int i=0;
	int step;
	unsigned long long nb_seq_ok=0;
	char* a_adapters[50]; // 50 sequences d'adapters max
	for (j=0;j<50;j++)
		a_adapters[j]=NULL;
	
	char *rcadapter=RevCompSeq(adapter);
	char adapter_fw_rv[MAXREADNT];
	sprintf(adapter_fw_rv,"%s %s",adapter,rcadapter);
	Split(adapter_fw_rv,' ',a_adapters);
	free(rcadapter);
//	fprintf(stderr,"indice nb_adapt_raw = %d\n",nb_adapt_raw);
	
	dejavucode_t *h_phix=NULL;
	HashAdapter(read_booster,phix,&h_phix);
	
	int nb_adapt = InitAdapterSetFromDataset(read_booster,a_adapters);
	
/*
	fprintf(stderr,"nb_adapt_edit = %d\n",nb_adapt);
	
	for(j=0;j<50;j++)
		if ( a_adapters[j] != NULL )
			fprintf(stderr,"%d %s\n",j,a_adapters[j]);
*/
	
	FILE *a_f_out[MAX_CPUS];
	for(j=0;j<read_booster->nb_cpus;j++)
		a_f_out[j]=GetStreamOut(lof_clean[j]);
	
	unsigned long long a_nb_seq_with_adaper[MAX_THREAD_LOCKS];
	unsigned long long a_nb_seq_trimmed[MAX_THREAD_LOCKS];
	unsigned long long a_nb_ovl_reads[MAX_THREAD_LOCKS];
	unsigned long long a_nb_seq_ok[MAX_THREAD_LOCKS];
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		a_nb_seq_with_adaper[i]=0;
		a_nb_seq_trimmed[i]=0;
		a_nb_ovl_reads[i]=0;
		a_nb_seq_ok[i]=0;
	}
	unsigned long long nb_seq_with_adapter=0;
	unsigned long long nb_seq_trimmed=0;
	unsigned long long nb_ovl_reads=0;
	unsigned long long nb_seq;
	int maxrdlen=read_booster->min_readlen_foruse;
	while(lof[idx] != NULL)
	{
		step=GetFormat(lof[idx]);
		nb_seq = GetNumberOfSequencePairs(seq2counts,lof[idx],read_booster->paired_files,read_booster->single_end,FALSE);
		fprintf(stderr,"File: %s\n",lof[idx]);
		
		arg_threads_t argt;
		argt.complex_arg[0]=read_booster;
		argt.complex_arg[1]=lof[idx];
		if ( read_booster->paired_files && ! read_booster->single_end )
		{
			fprintf(stderr,"+ %s\n",lof[idx+1]);
			argt.complex_arg[2]=lof[idx+1];
		}
		else
			argt.complex_arg[2]=NULL;
		argt.complex_arg[3]=&step;
		argt.complex_arg[4]=&a_adapters;
		argt.complex_arg[5]=&a_f_out;
		argt.complex_arg[6]=&maxrdlen;
		argt.complex_arg[7]=a_nb_seq_with_adaper;
		argt.complex_arg[8]=a_nb_seq_trimmed;
		argt.complex_arg[9]=a_nb_ovl_reads;
		argt.complex_arg[10]=a_nb_seq_ok;
		argt.complex_arg[11]=&h_phix;
		
		InitSMPStatusBar(read_booster->nb_cpus,&argt);
		Multithreading(CleanUpDatasetSlice,&argt,nb_seq,read_booster->nb_cpus);
		CloseSMPStatusBar(read_booster->nb_cpus,&argt);
		
		if ( read_booster->paired_files && ! read_booster->single_end )
			idx+=2;
		else
			idx++;
		CloseStatusBar();
	}
	for(j=0;j<read_booster->nb_cpus;j++)
		CloseStream(lof_clean[j],a_f_out[j]);
	
	FreeSplit(a_adapters,nb_adapt);
	DeleteDejaCode(&h_phix);
	
	for(i=0;i<read_booster->nb_cpus;i++)
	{
		nb_seq_with_adapter += a_nb_seq_with_adaper[i];
		nb_seq_trimmed += a_nb_seq_trimmed[i];
		nb_ovl_reads   += a_nb_ovl_reads[i];
		nb_seq_ok      += a_nb_seq_ok[i];
	}
	
	int coverage = (int)((nb_seq_ok*(unsigned long)maxrdlen*2)/read_booster->expected_nt_size);
	
	if ( read_booster->debug )
	{
		fprintf(stderr,"# reads with adapter/phix: %llu\n",nb_seq_with_adapter);
		fprintf(stderr,"# kmer-trimmed reads : %llu\n",nb_seq_trimmed);
		fprintf(stderr,"# inserts with overlapping reads : %llu\n",nb_ovl_reads);
		fprintf(stderr,"# pairs ok : %llu\n",nb_seq_ok);
		fprintf(stderr,"Coverage on cleaned dataset: %dx\n",coverage);
	}
	// lorsqu'il y a plus de 1 cpu, j'utilise cette fonction pour fusionner les resultats
	if ( coverage > 200 || read_booster->nb_cpus > 1 ) 
	{
		int step = (int)(coverage/100);
		if ( step == 0 )
			step=1;
		ResizeDatasetCoverage(lof_clean,cleanupdataset,step,read_booster->compress);
	}
	else
		strcpy(cleanupdataset,lof_clean[0]);
	if ( read_booster->nb_cpus > 1 )
	{
		char cmd[1000]="";
		for(i=0;i<read_booster->nb_cpus;i++)
		{
			sprintf(cmd,"\\rm -f %s\n",lof_clean[i]);
			int ret_code = system(cmd);
			if ( lof_clean[i] != NULL )
			{
				free(lof_clean[i]);
				lof_clean[i]=NULL;
			}
		}
	}
}

void *CleanUpDatasetSlice(void *argt)
{
	arg_threads_t *thr_data=(arg_threads_t *)argt;
	unsigned long long imin=thr_data->imin;
	unsigned long long imax=thr_data->imax;
	//fprintf(stderr,"imin=%ld, imax=%ld, thread_id=%d\n",thr_data->imin, thr_data->imax,thr_data->thread_id);
	read_booster_t *read_booster=(read_booster_t *)thr_data->complex_arg[0];
	char   *file1=(char *)thr_data->complex_arg[1];
	char   *file2=(char *)thr_data->complex_arg[2];
	int    step=*((int *)thr_data->complex_arg[3]);
	char  **a_adapters=((char **)thr_data->complex_arg[4]);
	FILE **a_f_out=((FILE **)thr_data->complex_arg[5]);
	int    maxrdlen=*((int *)thr_data->complex_arg[6]);
	unsigned long long *a_nb_seq_with_adaper=(unsigned long long *)thr_data->complex_arg[7];
	unsigned long long *a_nb_seq_trimmed=(unsigned long long *)thr_data->complex_arg[8];
	unsigned long long *a_nb_ovl_reads=(unsigned long long *)thr_data->complex_arg[9];
	unsigned long long *a_nb_seq_ok=(unsigned long long *)thr_data->complex_arg[10];
	dejavucode_t **h_phix=(dejavucode_t **)thr_data->complex_arg[11];
	
	FILE *f_out=a_f_out[thr_data->thread_id];
	
	unsigned long long decime=(imax-imin)/10;
	if ( decime == 0 )
		decime=1;
	
	char read1[MAXREADNT]="";
	char read2[MAXREADNT]="";
	char line[MAXLINE];
	char *rcread2;
	hsp_t hsp;
	
	FILE *f_reads  = GetStreamIn(file1);
	FILE *f_reads2 = NULL;
	if ( read_booster->paired_files )
		f_reads2 = GetStreamIn(file2); 
	int j;
	long long k=-1;
	unsigned long cpt=imin;
	char *sseq=NULL;
	int done=FALSE;
	short **p_mat=NULL;
	MallocMatrixOfShort(&p_mat,MAXREADNT_MATRIX);
	while((fgets(line,MAXLINE,f_reads) != NULL) && ! done)
	{
		k++;
		if ( k > imax )
		{
			done=TRUE;
			continue;
		}
		if (read_booster->single_end )
		{
			for(j=1;j<step;j++)
			{
				if ( fgets(line,MAXLINE,f_reads) == NULL )
					exit(EXIT_FAILURE);
				if (j==1)
				{
					strcpy(read1,line);
					read1[strlen(read1)-1]='\0';// c'est -1 a cause du \n 
				}
			}
		}
		else
		{
			for(j=1;j<step*2;j++)
			{
				if ( read_booster->paired_files && j >= step )
				{
					if ( fgets(line,MAXLINE,f_reads2) == NULL )
						exit(EXIT_FAILURE);
				}
				else if ( fgets(line,MAXLINE,f_reads) == NULL )
					exit(EXIT_FAILURE);
				if (j==1)
				{
					strcpy(read1,line);
					read1[strlen(read1)-1]='\0';// c'est -1 a cause du \n 
				}
				else if ( j == (1+step) )
				{
					strcpy(read2,line);
					read2[strlen(read2)-1]='\0';// c'est -1 a cause du \n
				}
			}
		}
		if ( k >= imin && k <= imax )
		{
			if ( (k-imin) % decime == 0 )
				SMPStatusBar(thr_data->thread_id,(unsigned long long)(k-imin),(unsigned long long)(imax-imin+1),thr_data);
			
			// maintenant que je prend toute la lecture, je peux tomber sur des N en fin qui me font tout rejeter
			// alors que c'est ok sur maxrdlen
			if ( (sseq=index(read1,'N')) != NULL )
				*sseq='\0';
			if ( strlen(read1) < maxrdlen )
				continue;
			
			if ( ! read_booster->single_end )
			{
				if ( (sseq=index(read2,'N')) != NULL )
					*sseq='\0';
				if ( strlen(read2) < maxrdlen )
					continue;
			}
			
			if ( TrimReadWithRareKmers(read_booster,read1) >= maxrdlen )
			{
				if ( RemoveAdapterByAlign(read1,a_adapters,'1',p_mat) >= maxrdlen && ! RemoveAdapterByKmer(read_booster,read1,h_phix)  )
				{
					if ( read_booster->single_end )
					{
						fprintf(f_out,">%ld/1\n%s\n",cpt,read1);
						cpt++;
						*(a_nb_seq_ok+thr_data->thread_id)+=1;
					}
					else
					{
						if ( TrimReadWithRareKmers(read_booster,read2) >= maxrdlen )
						{
							if ( RemoveAdapterByAlign(read2,a_adapters,'2',p_mat) >= maxrdlen && ! RemoveAdapterByKmer(read_booster,read2,h_phix) )
							{
								// l'alignement est tres strict, je dois en louper
									rcread2=RevCompSeq(read2);
									PairwiseMatchUngapped(read1,rcread2,&hsp,p_mat);
									free(rcread2);
									if ( (hsp.mm <= read_booster->maxmm || hsp.pci >= read_booster->minpci) &&  hsp.len >= read_booster->minovl && hsp.tstart==1 && hsp.qend == strlen(read1) )
									{
										//PrintHsp(&hsp);
										*(a_nb_ovl_reads+thr_data->thread_id)+=1;
									}
									else
									{
										fprintf(f_out,">%ld/1\n%s\n>%ld/2\n%s\n",cpt,read1,cpt,read2);
										cpt++;
										*(a_nb_seq_ok+thr_data->thread_id)+=1;
									}
							}
							else
								*(a_nb_seq_with_adaper+thr_data->thread_id)+=1;
						}
						else
							*(a_nb_seq_trimmed+thr_data->thread_id)+=1;
					}
				}
				else
					*(a_nb_seq_with_adaper+thr_data->thread_id)+=1;
			}
			else
				*(a_nb_seq_trimmed+thr_data->thread_id)+=1;
		}
	}
	FreeMatrixOfShort(&p_mat[0],MAXREADNT_MATRIX);
	CloseStream(file1,f_reads);
	if ( read_booster->paired_files && ! read_booster->single_end )
		CloseStream(file2,f_reads2);
	return(NULL);
}

// je ne fais pas du ramdom car je veux etre sur de prendre dans les clusters rares

int ResizeDatasetCoverage(lof_t lof,char *cleanaupdataset,int step,short compress)
{
	char newfile[MAXFILENAME];
	char ext[10]="";
	if ( compress )
		strcpy(ext,".gz");
	if ( step > 1 )
	{
		fprintf(stderr,"ResizeDatasetCoverage: step=%d\n",step);
		sprintf(newfile,"%s.one_over_%d%s",cleanaupdataset,step,ext);
	}
	else
		sprintf(newfile,"%s.merged%s",cleanaupdataset,ext);
	
	fprintf(stderr,"%s<\n",newfile);
	
	FILE *f_out=GetStreamOut(newfile);
	unsigned long long cpt=0;
	int idx=0;
	while(lof[idx] != NULL)
	{
		FILE *f_in=GetStreamIn(lof[idx]);
		char line[MAXLINE];
		int ok=FALSE;
		int stepcpt=0;
		// attention j'ai des paires
		while(fgets(line,MAXLINE,f_in) != NULL)
		{
			if ( line[0] == '>' )
			{
				if ( cpt % 4 == 0  )
				{
					if ( ++stepcpt == step )
					{
						ok = TRUE;
						stepcpt=0;
					}
					else
						ok = FALSE;
				}
				cpt++;
			}
			if ( ok )
				fputs(line,f_out);
		}
		CloseStream(lof[idx],f_in);
		idx++;
	}
	CloseStream(newfile,f_out);
	strcpy(cleanaupdataset,newfile);
	return(TRUE);
}

/*
 * 
 */