/*--------------------------------
 * 
 * Jerome.Gouzy@inra.fr
 * Created: 20140120
 * Copyright INRA, All rights reserved
 * 
 * $Id: generalbioinfo.h 516 2016-08-17 12:29:37Z gouzy $
 * 
 ---------------------------------*/

#ifndef GENERALBIOINFO_H

#define GENERALBIOINFO_H

#include "stdio.h"
#include "stdlib.h"
#include "errno.h"
#include "string.h"
#include "general.h"

typedef struct hsp
{
	float  mm;
	short  gap;
	double evalue;
	float  score;
	float  pci;
	char   query[100]; // id
	char   target[100];
	int    qstart;
	int    qend;
	int    tstart;
	int    tend;
	int    len;
	char   strand;
} hsp_t;

typedef unsigned int kmercode_t;

#define DROPOFF 5

char 	*RevCompSeq(char *str);
void 	ParseAALNLine(char *hit,int query_col,hsp_t *hsp);
int 	PairwiseMatchUngapped(char *query,char *target,hsp_t *hsp,short **p_mat);
int 	PairwiseMatchGapped(char *query,char *target,hsp_t *hsp,short gap_penalty,short **p_mat);
int PairwiseMatchGappedConcatTargetsForAssembly(char * query,char * p_targets, int tgtlen,hsp_t *hsp[],short gap_score,int * p_mat_block, int n_targets,int ref_dropoff,float minpci,int minovl,int maxovht);
void 	PrintHsp(hsp_t *hsp);
void 	InitHsp(hsp_t *hsp);
void 	AssertKmerEncoding(int klen,kmercode_t code);
static int KmerEncode(char *str,int klen,kmercode_t *code);
static int KmerEncodeOS(char *str,int klen,kmercode_t *code,kmercode_t mask);
static void KmerDecode(kmercode_t code,int klen,char *kmerdec);
static int KmerEncodeComplex(char *str,int klen,kmercode_t *code,kmercode_t mask);
static int KmerEncodeHub(char *str,int klen,kmercode_t *code,kmercode_t mask,int init);
static void RevCompCode(kmercode_t code,int klen, kmercode_t *rccode);

void MallocMatrixOfShort(short ***p_mat,int len); 	//MallocMatrixOfShort(&p_mat,MAXREADNT);
void FreeMatrixOfShort(short **p_mat,int len);      //FreeMatrixOfShort(&p_mat[0],MAXREADNT);
void MallocMatrixOfFloat(float ***p_mat,int x,int y);
void FreeMatrixOfFloat(float **p_mat,int x);
				
// 20151228: par rapport aux messages sur la vectorisation j'ai remplace les switchs et les comparaisons de char. J'ai aussi evite le return en cas de caractere inconnu
inline int KmerEncode(char *str,int klen,kmercode_t *code)
{
	//assert(str);
	*code=0;
	int i;
	int val;
	
	for(i=0;i<klen;i++)
	{
		val = *(str+i);
		if ( val > 84 )
			val -= 32; // je passe en majuscule
		if ( val == 65 ) // 'A'
		{
			;
		}
		else if ( val == 84 ) // 'T'
		{
			*code|=3;
		}
		else if ( val == 67 ) // 'C'
		{
			*code|=1;
		}
		else if ( val == 71 ) // 'G'
		{
			*code|=2;
		}
		else 
		{
			i=2*klen; // on ne peut pas vectoriser si on n'a pas un nombre connu d'iteration
		}
		if ( i < (klen-1) )
			*code <<= 2;
	}
	if ( i >= 2*klen )
	  return(FALSE);
	return(TRUE);
}


inline int KmerEncodeOS(char *str,int klen,kmercode_t *code,kmercode_t mask)
{
	*code &= mask;
	*code <<= 2;
	int val = *(str+klen-1);
	if ( val > 84 )
		val-=32;
	if ( val == 65 ) // 'A'
	{
		;
	}
	else if ( val == 84 ) // 'T'
	{
		*code|=3;
	}
	else if ( val == 67 ) // 'C'
	{
		*code|=1;
	}
	else if ( val == 71 ) // 'G'
	{
		*code|=2;
	}
	else 
	{
		return(FALSE);
	}
	return(TRUE);
}



inline void KmerDecode(kmercode_t code,int klen,char *kmerdec)
{
	int i;
	for(i=klen-1;i>=0;i--)
	{
		switch (code % 4)
		{
			case 0:
				kmerdec[i]='A';
				break;
			case 1:
				kmerdec[i]='C';
				break;
			case 2:
				kmerdec[i]='G';
				break;
			case 3:
				kmerdec[i]='T';
				break;
		}
		code >>= 2;
	}
	kmerdec[klen]='\0';
}

inline int KmerEncodeComplex(char *str,int klen,kmercode_t *code,kmercode_t mask)
{
	//assert(str);
	int i;
	kmercode_t code1=0;
	short offsetk=klen-15;
	if (! KmerEncode(str,15,&code1) )
		return(FALSE);
	kmercode_t code2=code1;
	for(i=0;i<offsetk;i++)
		if ( ! KmerEncodeOS(str+i,15,&code2,mask) )
			return(FALSE);
	*code=(code1+code2);
	return(TRUE);
}

inline int KmerEncodeHub(char *str,int klen,kmercode_t *code,kmercode_t mask,int init)
{
	if ( klen > 16 )
		return KmerEncodeComplex(str,klen,code,mask);
	else
	{
		if ( ! init )
			return KmerEncode(str,klen,code);
		else
			return KmerEncodeOS(str,klen,code,mask);
	}
}


// 20151228: remplace le switch par a_ACGT_rccode;
static int a_ACGT_rccode[4]={3,2,1,0};
inline void RevCompCode(kmercode_t code,int klen, kmercode_t *rccode)
{
	int i;
	*rccode=0;
	for(i=klen-1;i>=0;i--)
	{
		*rccode |= a_ACGT_rccode[code%4];
		/*
		switch (code % 4)
		{
			case 0:
				*rccode |= 3;
				break;
			case 1:
				*rccode |= 2;
				break;
			case 2:
				*rccode |= 1;
				break;
			case 3:
		//		*rccode |= 0;
				break;
		}
		*/
		code >>= 2;
		if ( i > 0 )
		{
			*rccode <<= 2;
		}
	}
}

#endif